/*
    clxmms - Command-Line Xmms controlling program
    Copyright (C) 2001  Krzysztof Garus <kgarus@bigfoot.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <xmmsctrl.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <wordexp.h>
#include <regex.h>
#include <signal.h>
#include <glib.h>
#include <unistd.h>
#include <errno.h>

// VOLUME & BALANCE procent change:
#define VOL_PROC 10
#define VOL_PROC2 20

#if 1
#define CLBOLD   "\033[1m"
#define CLNORMAL "\033[0m"
#define CLRED    "\033[31m"
#define CLGREEN  "\033[32m"
#define CLBROWN  "\033[33m"
#define CLBLUE   "\033[34m"
#define CLYELLOW CLBOLD CLBROWN
#define CLLTRED  CLBOLD CLRED
#else
#define CLBOLD   ""
#define CLNORMAL ""
#define CLRED    ""
#define CLGREEN  ""
#define CLBROWN  ""
#define CLBLUE   ""
#define CLYELLOW ""
#endif

char *helptext =
     "  exec [display]\n"
     "  quit\n"
     "  play\n"
     "  stop\n"
     "  next\n"
     "  prev\n"
     "  shedule [-c] [-i pos] [num ...]\n"
     "  add file ...\n"
     "  del [-d] num ...\n"
     "  pos [num]\n"
     "  list [-B before] [-A after] [-C before-and-after]\n"
	 "       [-F from] [-T to]        (negative means \"from end\")\n"
     "  status\n"
     "  find [-i] pattern\n"
     "  loop\n"
     "  volume\n"
     "	down\n"
     "  up\n"
     "  left\n"
     "  right\n"
     "  shuffle\n"
	 "  ls\n"
	 "  pwd\n"
	 "  cd\n";


int session = 0,
    prev_pos = -1,
	color = 1;

GArray *shedlist = NULL;

typedef enum
{
	CMD_PLAY = 0,
	CMD_STOP,
	CMD_PAUSE,
	CMD_NEXT,
	CMD_PREV,
	CMD_LIST,
	CMD_EXEC,
	CMD_QUIT,
	CMD_EXIT,
	CMD_POS,
	CMD_SHEDULE,
	CMD_ADD,
	CMD_DEL,
	CMD_FIND,
	CMD_HELP,
	CMD_STATUS,
	CMD_LOOP,
	CMD_SHUFFLE,
	CMD_LS,
	CMD_CD,
	CMD_PWD,
	/*CMD_SESSION,*/
	CMD_GET_VOL,
	CMD_SET_VOL_UP,		
	CMD_SET_VOL_DN,		
	CMD_SET_VOL_R,		
	CMD_SET_VOL_L,		
	CMD_UNKNOWN,
	CMD__END,
	CMD_A_PART
} command_t;

const char *command[] =
{
	[CMD_PLAY] = "play",
	[CMD_STOP] = "stop",
	[CMD_PAUSE] = "pause",
	[CMD_NEXT] = "next",
	[CMD_PREV] = "prev",
	[CMD_LIST] = "list",
	[CMD_EXEC] = "exec",
	[CMD_QUIT] = "quit",
	[CMD_EXIT] = "exit",
	[CMD_POS] = "pos",
	[CMD_SHEDULE] = "shedule",
	[CMD_ADD] = "add",
	[CMD_DEL] = "del",
	[CMD_FIND] = "find",
	[CMD_HELP] = "help",
	[CMD_STATUS] = "status",
	[CMD_LOOP] = "loop",
	[CMD_SHUFFLE] = "shuffle",
	[CMD_LS] = "ls",
	[CMD_CD] = "cd",
	[CMD_PWD] = "pwd",
	[CMD_GET_VOL] = "volume",
	[CMD_SET_VOL_UP] = "up",
	[CMD_SET_VOL_DN] = "down",
	[CMD_SET_VOL_R] = "right",
	[CMD_SET_VOL_L] = "left",
	/*[CMD_SESSION] = "session",*/
	[CMD__END] = NULL

};

command_t GetCommand(char *cmdstr)
{
	int i, len, nmatch;
	command_t cmd;
	
	if (cmdstr == NULL)
		return CMD_UNKNOWN;
	
	len = strlen(cmdstr);
	nmatch = 0;
	for (i=0; command[i]; i++)
	{
		if (strncmp(command[i], cmdstr, len) == 0)
		{
			cmd = i;
			nmatch++;
		}
	}

	if (nmatch == 1)
		return cmd;
	else if (nmatch == 0)
		return CMD_UNKNOWN;
	
	/* nmatch > 1 */
	fprintf(stderr, "Matched commands: ");
	for (i=0; command[i]; i++)
		if (strncmp(command[i], cmdstr, len) == 0)
			fprintf(stderr, "\"%s\" ", command[i]);
	fprintf(stderr, "(type \"help\" for menu)\n");

	return CMD_A_PART;
}


/* signals handling */
int got_exit_signal = 0;

void signal_h(int sig)
{
	got_exit_signal = 1;
}

void Do_List(int argc, char **argv)
{
	int nstart, nend, i, narg, tmp, len, pos;
	char *title, c;

	len = xmms_remote_get_playlist_length(session);
	pos = xmms_remote_get_playlist_pos(session);
	
	nstart = 0;
	nend = len;

	optind = 1;
	while ((c = getopt(argc, argv, "A:B:C:F:T:")) != -1)
	{
		if (c != '?')
			narg = strtol(optarg, NULL, 10);
		switch(c)
		{
			case 'B': nstart = pos - narg; break;
			case 'A':   nend = pos + narg+1; break;
			case 'C': nstart = pos - narg;
			            nend = pos + narg+1; break;
			case 'F':
				if (narg > 0)
					nstart = narg - 1;     /* -1 because human starts from 1 */
				else
					nstart = len - (-narg) -1 +1;
				break;
			case 'T':
				if (narg > 0)
					nend = narg - 1 + 1;
				else
					nend = len - (-narg) -1 +1 +1;
				break;
			default:
				fprintf(stderr, "Usage: list [-A before] [-B after] [-C before-and-after] [-F from] [-T to]\n");
				return;
		}
	}

	if (nend <= nstart)  /* reverse range */
	{
		tmp = nend;
		nend = nstart+1;
		nstart = tmp-1;
	}
	if (nstart < 0)     nstart = 0;
	if (nend < 0)       nend = 0;
	if (nstart > len)   nstart = len;
	if (nend > len)     nend = len;
	/*printf("List from=%d to=%d\n", nstart+1, nend);*/
	
	for (i=nstart; i<nend && !got_exit_signal; i++)
	{
		title = xmms_remote_get_playlist_title(session, i);
		if (title)
		{
			printf(color ? CLGREEN "%4d %s%s\n" CLNORMAL : "%4d %3$s\n", i+1,
			       pos == i ? CLYELLOW : CLNORMAL,
			       title);
			free(title);
		}
	}
}


void Do_Find(int argc, char **argv)
{
	int len, pos, i;
	char *re;
	regex_t preg;
	char *title;
	int icase = 0;

	if (argc == 1)
	{
		fprintf(stderr, "Usage: find [-i] <regexp>\n");
		return;
	}
	
	if (strcmp(argv[1], "-i") == 0)
	{
		icase = REG_ICASE;
		argc--;
		argv++;
	}
	
	len = xmms_remote_get_playlist_length(session);
	pos = xmms_remote_get_playlist_pos(session);
	
	re = argv[1];
	if (regcomp(&preg, re, icase|REG_NOSUB|REG_EXTENDED))
		fprintf(stderr, "Bad regular expression!\n");
	else
		for (i=0; i<len && !got_exit_signal; i++)
		{
			title = xmms_remote_get_playlist_title(session, i);
			if (regexec(&preg, title, 0, NULL, 0) == 0)
				printf(color ? CLGREEN "%4d %s%s\n" CLNORMAL : "%4d %3$s\n", i+1,
				       pos == i ? CLYELLOW : CLNORMAL,
				       title);
			free(title);
		}
	regfree(&preg);
}


void XmmsStatus()
{
	int     pos = xmms_remote_get_playlist_pos(session),
	        len = xmms_remote_get_playlist_length(session);
	char *title = xmms_remote_get_playlist_title(session, pos);
	char *file = xmms_remote_get_playlist_file(session, pos);

	printf("\n Session: %d\n Playing: %s\n  Paused: %s\n"
	       "   Files: \%d\nPosition: \%d \"%s\" %s\n\n",
	       session,
	       xmms_remote_is_playing(session) ? "yes" : "no",
	       xmms_remote_is_paused(session) ? "yes" : "no",
	       len, pos+1, title ? title : "<no sound>",
		   file ? file : "");
	free(title);
}


int cmpr_int(int *i1, int *i2)
{
	return *i1 - *i2;
}

int DoCommandArgs(int argc, char **argv)
{
	int i,
	    num, *tab,
		len,
	    unlink_file;
	char path[PATH_MAX],
	     *path2,
	     *home,
		 *str, *str2,
	     *new_display;
	GList *list;
	command_t cmd;
	
	switch (cmd = GetCommand(argv[0]))
	{
		case CMD_EXEC:
			if (argc > 1)
				new_display = argv[1];
			else
				new_display = getenv("DISPLAY");
			if (!new_display)
				new_display = ":0";
			printf("Running xmms on display %s\n", new_display);
			
			if (!fork())
			{
				home = getenv("HOME");
				setpgrp();
				signal(SIGHUP, SIG_IGN);
				chdir(home ? home : "/");
				setenv("DISPLAY", new_display, 1);
				
				fclose(stdin);
				fclose(stdout);
				fclose(stderr);
				execlp("xmms", "xmms", NULL);
			}
		break;

		case CMD_PLAY:  xmms_remote_play(session); break;
		case CMD_QUIT:  xmms_remote_quit(session); break;
		case CMD_STOP:  xmms_remote_stop(session); break;
		case CMD_PAUSE: xmms_remote_pause(session); break;
		case CMD_NEXT:  xmms_remote_playlist_next(session); break;
		case CMD_PREV:  xmms_remote_playlist_prev(session); break;
		case CMD_LOOP:  xmms_remote_toggle_repeat(session); break;
		case CMD_SHUFFLE:  xmms_remote_toggle_shuffle(session); break;
		case CMD_STATUS: XmmsStatus(); break;

		int vol=0; //magic:
		case CMD_SET_VOL_UP:	vol=VOL_PROC2;
		case CMD_SET_VOL_DN:	vol-=VOL_PROC;
		case CMD_GET_VOL: 
			       {
				int vl,vr,bt;				      
			       xmms_remote_get_volume(session, &vl, &vr);
			       vl+=vol;
			       vr+=vol;
			       xmms_remote_set_volume(session, vl, vr);
			       xmms_remote_get_volume(session, &vl, &vr);
			       bt=xmms_remote_get_balance(session);
			       printf("Left:%d Right:%d Balance: %d\n", vl, vr,bt);
			       }
				break;
		int bal=0;				 
		case CMD_SET_VOL_R:	bal=VOL_PROC2;
		case CMD_SET_VOL_L:	bal-=VOL_PROC;
					{
						int bt=0;
						bt=xmms_remote_get_balance(session);
						bt+=bal;
						xmms_remote_set_balance(session,bt);
						bt=xmms_remote_get_balance(session);
						printf("Balance: %d\n", bt);
					}
			break;
		case CMD_POS:
			if (argc == 2)
				xmms_remote_set_playlist_pos(session, strtol(argv[1], NULL, 10) - 1);
			else
				printf("Pos: %d\n", xmms_remote_get_playlist_pos(session) + 1);
		break;
		
		case CMD_SHEDULE:
			if (argc == 1)
			{
				if (shedlist->len == 0)
					printf("No songs sheduled\n");
				for (i=0; i<shedlist->len; i++)
				{
					int n = g_array_index(shedlist, int, i);
					char *title = xmms_remote_get_playlist_title(session, n);
					if (title)
					{
						printf(color ? CLYELLOW "%2d " CLGREEN "%4d %s%s\n" CLNORMAL : "%2d %4d %3$s\n",
								i+1, n+1,
						       /*pos == i ? CLYELLOW :*/ CLNORMAL,
						       title);
						free(title);
					}
					else // not found?
						g_array_remove_index(shedlist, i);
				}
			}
			else if (strcmp(argv[1], "-c") == 0)
			{
				g_array_set_size(shedlist, 0);
				printf("Shedule list cleared\n");
			}
			else
			{
				int arg1=1;
				int insnum = shedlist->len;
				int nadd = 0;
				char *p;
				len = xmms_remote_get_playlist_length(session);
				if (strcmp(argv[1], "-i") == 0)
				{
					insnum = strtol(argv[2], &p, 10)-1;
					if (*p != '\0' || insnum <0 || insnum >= len)
						break;
					arg1 += 2;
				}
				for (i=arg1; i<argc; i++)
				{
					int n = strtol(argv[i], &p, 10) - 1;
					if (*p == '\0' && n >=0 && n < len)
					{
						g_array_insert_val(shedlist, insnum++, n);
						nadd++;
					}
				}
				printf("Sheduled %d songs, total: %d\n", nadd, shedlist->len);
			}
		break;
		
		case CMD_ADD:
			getcwd(path, PATH_MAX);
			list = NULL;
			for (i=1; i<argc; i++)
			{
				if (argv[i][0] == '/')
					path2 = strdup(argv[i]);
				else
					path2 = g_strdup_printf("%s/%s", path, argv[i]);
				/*printf("path = \"%s\"\n", path2);*/
				list = g_list_append(list, path2);
			}
			xmms_remote_playlist_add(session, list);
			g_list_free(list);
		break;
		
		case CMD_DEL:
			if (argc > 1 && strcmp(argv[1], "-d") == 0)
			{
				unlink_file = 1;
				argv++;
				argc--;
			}
			else
				unlink_file = 0;
			
			tab = (int *)malloc(argc * sizeof(int));
			if (argc == 1)
			{
				num = 1;
				tab[0] = xmms_remote_get_playlist_pos(session);
			}
			else
			{
				num = 0;
				for (i=1; i<argc; i++)
					tab[num++] = strtol(argv[i], NULL, 10) - 1;
			}
			qsort(tab, num, sizeof(int), (int(*)(const void *,const void *))cmpr_int);
			for (i=num-1; i>=0; i--)
			{
				if (unlink_file)
				{
					char *file = xmms_remote_get_playlist_file(session, tab[i]);
					
					if (!file)
						continue;
					printf("Deleting: \"%s\": ", file);
					if (unlink(file))
						printf(color ? CLRED "%s" CLNORMAL "\n" : "%s\n", sys_errlist[errno]);
					else
						printf(color ? CLGREEN "OK" CLNORMAL "\n" : "OK\n");
						
					free(file);
				}
				xmms_remote_playlist_delete(session, tab[i]);
			}
			if (tab[0] <= xmms_remote_get_playlist_pos(session))
				prev_pos = -1; /* force to redisplay new song */
			free(tab);
		break;
		
		case CMD_LS:
			str = g_strjoinv(" ", argv+1);
			str2 = g_strdup_printf("/bin/ls --color=tty %s", str);
			system(str2);
			free(str);
			free(str2);
		break;
		
		case CMD_CD: if (argc == 2) chdir(argv[1]); break;
		case CMD_PWD:
			getcwd(path, PATH_MAX);
			printf("%s\n", path);
		break;

		case CMD_LIST: Do_List(argc, argv); break;
		case CMD_FIND: Do_Find(argc, argv); break;
		case CMD_A_PART: break; /* user was told */
		case CMD_HELP:
		case CMD_UNKNOWN:
		
		default:
			printf("%s", helptext);
			return cmd == CMD_HELP ? EXIT_SUCCESS : EXIT_FAILURE;
	};
	
	return EXIT_SUCCESS;
}


/*
 * Some global variables
 */
char *oldcmd = NULL;
int break_main_loop = 0,
    timeout_sec = 1;

void readline_callback(char *cmd)
{
	wordexp_t we;

	if (cmd == NULL || !strcmp(cmd, "exit") || got_exit_signal)
	{
		break_main_loop = 1;
		rl_callback_handler_remove();
		return;
	}
	
	if (wordexp(cmd, &we, WRDE_SHOWERR))
		fprintf(stderr, "Bad command line\n");
	else if (we.we_wordc > 0)
		DoCommandArgs(we.we_wordc, we.we_wordv);
	
	wordfree(&we);
	if (!oldcmd || strcmp(cmd, oldcmd))
		add_history(cmd);
	free(oldcmd);
	oldcmd = cmd;

	timeout_sec = 1;
}


void rl_printf(char *fmt, ...)
{
	int i, old_end, len;
	char *old_prompt;
	va_list ap;
	
	old_end = rl_end;
	old_prompt = rl_prompt;
	len = strlen(old_prompt);
	rl_end = 0;
	rl_prompt = "";
	rl_redisplay();
	printf("\r");
	for (i=0; i<len; i++)
		putchar(' ');
	printf("\r");

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
	
	rl_end = old_end;
	rl_prompt = old_prompt;
	rl_forced_update_display();
}


/*
 * Completition
 */

char *gen_command_name(const char *text, int state)
{
	static int i, len;

	if (!state)
	{
		i = 0;
		len = strlen(text);
	}

	for (; command[i]; i++)
		if (strncmp(command[i], text, len) == 0)
			return strdup(command[i++]);
	
	return NULL;
}

char ** cmd_compl(const char *text, int start, int end)
{
	char **matches;

	matches = NULL;
	if (start==0)
		matches = rl_completion_matches(text, gen_command_name);
		
	return matches;
}

int main(int argc, char *argv[])
{
	char history[PATH_MAX] = "\0",
	     *home;
	int pos = 0;
	static int shedtests = 0;
	
	shedlist = g_array_new(FALSE, FALSE, sizeof(int));
	
	signal(SIGCHLD, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, signal_h);
	signal(SIGTERM, signal_h);

	/* basic options */
	if (argc>1 && (!strcmp(argv[1], "-h") || !strcmp(argv[1], "--help")))
	{
		printf("clxmms [-h|--help] [-v|--version] [command [args ...]]\n"
		      " Where command may be one of:\n\n%s\n", helptext);
		exit(0);
	}
	else if (argc>1 && (!strcmp(argv[1], "-v") || !strcmp(argv[1], "--version")))
	{
		printf("clxmms version %s\n", VERSION);
		exit(0);
	}
	
	/* make output colored? */
	color = isatty(1);
	
	/* batch mode */
	if (argc > 1)
		exit( DoCommandArgs(argc-1, argv+1) );
	
	/* history reading */
	home = getenv("HOME");
	snprintf(history, sizeof(history), "%s/.clxmms_history", home ? home : ".");
	read_history(history);
	if (history_length)
		oldcmd = strdup(history_get(history_length)->line);
	
	/* main loop */
	rl_callback_handler_install(
		color ? CLBOLD CLBLUE "XMMS" CLNORMAL "> " : "XMMS> ", readline_callback);
	rl_attempted_completion_function = cmd_compl;
	while (!break_main_loop)
	{
		fd_set fd;
		struct timeval tv;
		char *title;
		
		FD_ZERO(&fd);
		FD_SET(0, &fd);
		tv.tv_sec = timeout_sec;
		tv.tv_usec = 0;

		switch(select(1, &fd, NULL, NULL, &tv))
		{
			case -1:
				if (errno != EINTR)
				{
					perror("select");
					break_main_loop = 1;
				}
			break;

			case 0: /* timeout */
				timeout_sec = 1;
				pos = xmms_remote_get_playlist_pos(session);
				if (pos != prev_pos)
				{
					gboolean sh = FALSE;
					
					if (shedlist->len)
					{
						int req = g_array_index(shedlist, int, 0);
						if (pos == req)
						{
							g_array_remove_index(shedlist, 0);
							sh = TRUE;
							shedtests = 0; /* reset */
						}
						else
						{
							xmms_remote_set_playlist_pos(session, req);
							if (++shedtests == 3)
							{
								title = xmms_remote_get_playlist_title(session, req);
								rl_printf(color ? CLRED "Can't play: " CLGREEN "%4d " CLYELLOW "%s" CLNORMAL "\n"
								                : "Can't play: %4d %s\n", req+1, title ?: "<title not available>");
								free(title);
								g_array_remove_index(shedlist, 0);
								shedtests = 0;
							}
							timeout_sec = 3;
							break; // don't show now, but later
						}
					}
					
					title = xmms_remote_get_playlist_title(session, pos);
					if (title)
					{
						rl_printf(color ? "%sNow: " CLGREEN "%4d " CLYELLOW "%s" CLNORMAL "\n"
						                : "Now%2$s: %43$d %s\n",
										sh ? CLLTRED:CLBROWN, pos+1, title);
						free(title);
					}
					
					prev_pos = pos;
				}
				/*timeout_sec = 5;*/
			break;
				
			default:
				if (FD_ISSET(0, &fd))
					rl_callback_read_char();
			
		};
		
		if (got_exit_signal)
			break;
	}
	
	printf("\n");
	write_history(history);
	return 0;
} /* main */


// This must be last line !
// vi:syntax=c:tw=78:ts=3:sw=4
